// Run a bunch of filters in a chain, pipeing one to the next. Return
// an error status if any of the commands in the pipe returns an error
// status.
//
// Commands and command arguments are passed thusly:
//
// ./dannyboy --prog=program --arg=arg1 --arg=arg2 --prog=program ...
//
// The --prog arg names a program, any following --arg arguments
// name arguments to be passed to the previous --prog, another --prog
// start a new program.
//
// The arg strings are passed in the argv array to the program, so be sure
// each string is intended to be a separate argument.
//
// Written by Tom Horsley, December 2011, released to the public domain
// for any use.

#include <string>
#include <vector>
#include <iostream>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

class OneProg {
public:

   OneProg(std::string init_progname);

   void
   AddArg(std::string new_arg);

   void
   SetInAndOut(int init_stdin, int init_stdout);

   void
   StartProgram();

   pid_t
   ProgramPid() {
         return program_pid;
      }

private:

   std::string progname;
   std::vector<std::string> args;

   int prog_stdin;
   int prog_stdout;

   pid_t program_pid;
};

static std::vector<OneProg *> progs;

static std::vector<int> pipes_to_close;

static void
close_all_pipes() {
   for (int c = 0; c < pipes_to_close.size(); ++c) {
      close(pipes_to_close[c]);
   }
}

#define IN_FD 0
#define OUT_FD 1

#define READ_END 0
#define WRITE_END 1

int
main(int argc, char ** argv) {
   OneProg * curprog = NULL;
   for (int i = 1; i < argc; ++i) {
      if (strncmp(argv[i], "--prog=", 7) == 0) {
         curprog = new OneProg(std::string(argv[i]+7));
         progs.push_back(curprog);
      } else if (strncmp(argv[i], "--arg=", 6) == 0) {
         if (curprog == NULL) {
            std::cerr << "--arg must follow --prog" << std::endl;
            return 2;
         } else {
            curprog->AddArg(std::string(argv[i]+6));
         }
      } else {
         std::cerr << "Unrecognized argument: " << argv[i] << std::endl;
         return 2;
      }
   }
   if (progs.size() <= 1) {
      std::cerr << "Must have more than one --prog argument." << std::endl;
      return 2;
   }

   // Build the pipe chain for all the programs

   int next_prog_stdin = IN_FD;
   int len = progs.size();
   for (int p = 0; p < len; ++p) {
      if (p == (len-1)) {

         // This is the last program, it gets the previous program's
         // stdout as its stdin, and it gets the normal stdout as its
         // stdout.

         progs[p]->SetInAndOut(next_prog_stdin, OUT_FD);
      } else {

         // This is the beginning or middle of the pipe chain. Use stdout
         // from previous program as stdin (which is really stdin for 1st
         // program), and make a new pipe for stdout (and to use as next
         // progs's stdin).

         int fds[2];
         if (pipe(fds) == -1) {
            const char * errmsg = strerror(errno);
            std::cerr << "Unable to create pipe: " << errmsg << std::endl;
            return 2;
         }
         pipes_to_close.push_back(fds[0]);
         pipes_to_close.push_back(fds[1]);
         progs[p]->SetInAndOut(next_prog_stdin, fds[WRITE_END]);
         next_prog_stdin = fds[READ_END];
      }
   }

   // Now start all the programs.

   for (int p = 0; p < len; ++p) {
      progs[p]->StartProgram();
   }

   // Close all the pipe file descriptors I'm holding open

   close_all_pipes();

   // And wait for all the programs to finish, taking note of any error
   // exit status that shows up.

   int exit_status = 0;
   int kidcount = progs.size();
   while (kidcount > 0) {
      int kidstat;
      pid_t wait_pid = waitpid((pid_t)-1, &kidstat, 0);
      if (wait_pid == (pid_t)-1) {
         if (errno == EINTR) {
            continue;
         } else {
            const char * errmsg = strerror(errno);
            std::cerr << "Error from waitpid: " << errmsg << std::endl;
            return 2;
         }
      }
      bool found_kid = false;
      for (int p = 0; (p < len) && (! found_kid); ++p) {
         if (progs[p]->ProgramPid() == wait_pid) {
            found_kid = true;
            if (! (WIFEXITED(kidstat) && (WEXITSTATUS(kidstat) == 0))) {
               if (exit_status == 0) {

                  // Remember exit status of first kid that exits abbynormally

                  if (WIFEXITED(kidstat)) {
                     exit_status = WEXITSTATUS(kidstat);
                  } else {
                     exit_status = 255;
                  }
               }
            }
         }
      }
      if (! found_kid) {
         std::cerr << "Why did I get status for a child ("
                   << (int)wait_pid
                   << ") I didn't start?" << std::endl;
         return 2;
      } else {
         --kidcount;
      }
   }
   return exit_status;
}

OneProg::OneProg(std::string init_progname)
 : progname(init_progname), prog_stdin(-1), prog_stdout(-1) {
}

void
OneProg::AddArg(std::string new_arg) {
   args.push_back(new_arg);
}

void
OneProg::SetInAndOut(int init_stdin, int init_stdout) {
   prog_stdin = init_stdin;
   prog_stdout = init_stdout;
}

void
OneProg::StartProgram() {
   program_pid = fork();
   if (program_pid == (pid_t)-1) {
      const char * errmsg = strerror(errno);
      std::cerr << "fork() failed: " << errmsg << std::endl;
      exit(2);
   }
   if (program_pid != (pid_t)0) {
      return;
   }

   // OK, this is the child, I need to setup my stdin and stdout
   // correctly and exec the program with the given arguments.

   if (prog_stdin != IN_FD) {
      dup2(prog_stdin, IN_FD);
   }
   if (prog_stdout != OUT_FD) {
      dup2(prog_stdout, OUT_FD);
   }

   // Close all pipe descriptors now that I've duped mine

   close_all_pipes();

   // Now build argv array to exec the program with.

   char ** argv = new char * [args.size() + 2];

   argv[0] = (char *)progname.c_str();
   for (int i = 0; i < args.size(); ++i) {
      argv[i+1] = (char *)args[i].c_str();
   }
   argv[args.size() + 1] = NULL;
   execvp(argv[0], argv);
   const char * errmsg = strerror(errno);
   std::cerr << "exec of " << progname << " failed: " << errmsg << std::endl;
   exit(2);
}
