// trollfilter - mark troll email threads using a database of known trolls
// and troll message IDs. Reads mail on stdin, writes to stdout, adds the
// X-TrollFilter: header with the value "Yes" or "No"
//
// Written by Tom Horsley, December 2011, released to the public domain
// for any use.

#include "mail-message.h"
#include <iostream>
#include <sstream>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <string>
#include <vector>
#include <sqlite3.h>
#include <unistd.h>

static std::string database_name;
static std::vector<std::string> trolls;
static std::vector<std::string> untrolls;
static std::vector<std::string> droppings;

static void database_help();
static void troll_help();
static void untroll_help();
static void dropping_help();
static void clean_help();
static void help();

enum ProgramActions {
   PA_UNSPECIFIED,
   PA_MAINT,            // one of --troll, --untroll, or --dropping
   PA_TROLLADDR,        // --trolladdr
   PA_UNTROLLADDR,      // --untrolladdr
   PA_TROLLID,          // --trollid
   PA_CLEAN,            // --clean
   PA_FILTER,           // --filter
   PA_TEST,             // --test
   PA_LIST,             // --list
};

static ProgramActions what_to_do = PA_UNSPECIFIED;

static const char *
what_name(ProgramActions act) {
   switch(act) {
   case PA_MAINT:       return "--troll, --untroll, or --dropping";
   case PA_TROLLADDR:   return "--trolladdr";
   case PA_UNTROLLADDR: return "--untrolladdr";
   case PA_TROLLID:     return "--trollid";
   case PA_CLEAN:       return "--clean";
   case PA_FILTER:      return "--filter";
   case PA_TEST:        return "--test";
   case PA_LIST:        return "--list";
   default:             return "unknown";
   }
}

static void
say_what(ProgramActions new_action) {
   if ((what_to_do == new_action) || (what_to_do == PA_UNSPECIFIED)) {
      what_to_do = new_action;
      return;
   }
   std::cerr << "Option " << what_name(new_action)
             << " cannot be used with "
             << what_name(what_to_do) << std::endl;
   exit(2);
}

static const char initialize_sql [] =
   "create table threads ( threadNum INTEGER PRIMARY KEY, recent DATE );"
   "create table trolls ( email VARCHAR(50) PRIMARY KEY );"
   "create table droppings ( messageID VARCHAR(100) PRIMARY KEY, threadNum INTEGER );";

static sqlite3 * db = NULL;

static std::string
to_lower(std::string str) {
   std::string rval;
   int len = str.size();
   for (int i = 0; i < len; ++i) {
      char c = tolower(str[i]);
      rval.append(1,c);
   }
   return rval;
}

static void
add_one_troll(std::string the_troll) {
   std::string cmd("insert or ignore into trolls (email) values ('");
   cmd.append(to_lower(the_troll));
   cmd.append("');");
   char * errmsg;
   if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Adding troll error: " << errmsg << std::endl;
      sqlite3_free(errmsg);
   }
}

static void
delete_one_troll(std::string the_troll) {
   std::string cmd("delete from trolls where email='");
   cmd.append(to_lower(the_troll));
   cmd.append("';");
   char * errmsg;
   if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Removing troll error: " << errmsg << std::endl;
      sqlite3_free(errmsg);
   }
}

static bool
is_a_troll(std::string the_troll) {
   bool rval = false;
   std::string cmd("select * from trolls where email='");
   cmd.append(to_lower(the_troll));
   cmd.append("';");
   sqlite3_stmt * sp = NULL;
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      if (sqlite3_step(sp) == SQLITE_ROW) {

         // Finding a matching row means that the email address is indeed
         // a troll. Return true.

         rval = true;
      }
   }
   sqlite3_finalize(sp);
   return rval;
}

static void close_database() {
   if (db == NULL) {
      return;
   }

   char * errmsg;

   // Commit the transaction.

   if (sqlite3_exec(db, "commit;", NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Error committing transaction: "
                << errmsg
                << std::endl;
      sqlite3_free(errmsg);
   }

   sqlite3_close(db);
   db = NULL;
}

static void open_database() {
   if (db != NULL) {
      // Already opened database.
      return;
   }

   // Remember if the database existed before we opened it.

   bool need_to_init = access(database_name.c_str(), R_OK|W_OK) != 0;

   if (sqlite3_open_v2(database_name.c_str(),
                       &db,
                       SQLITE_OPEN_READWRITE|SQLITE_OPEN_CREATE,
                       NULL)) {
      std::cerr << "Cannot open database \""
                << database_name
                << "\": "
                << sqlite3_errmsg(db)
                << std::endl;
      sqlite3_close(db);
      exit(2);
   }

   char * errmsg;

   // Don't get busy errors if multiple runs happen to occur at the same
   // time, just wait up to a couple of seconds (which should be enough for
   // any one run to finish).

   sqlite3_busy_timeout(db, 2000);

   // Start a transaction that covers everything trollfilter does during
   // this run.

   if (sqlite3_exec(db, "begin;", NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Error starting transaction: "
                << errmsg
                << std::endl;
      sqlite3_free(errmsg);
   }

   if (need_to_init) {
      if (sqlite3_exec(db, initialize_sql, NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Unable to create required tables: "
                   << errmsg
                   << std::endl;
         sqlite3_free(errmsg);
         close_database();
         exit(2);
      }
   }
}

static bool
check_one_thread(std::string mid, sqlite3_int64 & thread_num) {
   bool rval = false;
   std::string cmd("select (threadNum) from droppings where messageID='");
   cmd.append(mid);
   cmd.append("';");
   sqlite3_stmt * sp = NULL;
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      if (sqlite3_step(sp) == SQLITE_ROW) {

         // We found the message ID in the troll droppings, need to
         // extract the associated thread number.

         rval = true;
         thread_num = sqlite3_column_int64(sp, 0);
      }
   }
   sqlite3_finalize(sp);
   return rval;
}

static void
record_troll_thread(
   std::string mid,
   bool has_thread_num,
   sqlite3_int64 thread_num) {

   std::stringstream ss;
   if (! has_thread_num) {

      // Apparently this is a new thread, we need to create a new
      // thread number for it.

      std::string cmd("insert into threads values(NULL,DATETIME('NOW'));");
      char * errmsg;
      if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Adding thread error: " << errmsg << std::endl;
         sqlite3_free(errmsg);
      }
      thread_num = sqlite3_last_insert_rowid(db);
      ss << thread_num;
   } else {

      // This is an existing thread, we should update the recent
      // activity date for the thread.

      ss << thread_num;
      std::string
         cmd("update threads set recent=DATETIME('NOW') where threadNum=");
      cmd.append(ss.str());
      cmd.append(";");
      char * errmsg;
      if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Updating thread error: " << errmsg << std::endl;
         sqlite3_free(errmsg);
      }
   }

   // Now add the new message ID to the droppings table with the
   // thread number.

   std::string cmd("insert or ignore into droppings values('");
   cmd.append(mid);
   cmd.append("',");
   cmd.append(ss.str());
   cmd.append(");");
   char * errmsg;
   if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Adding message ID error: " << errmsg << std::endl;
      sqlite3_free(errmsg);
   }
}

static bool
filter_message(bool do_test) {

   // Read mail message from cin and get header

   MailMessage mm;

   bool is_troll_thread = false;
   bool has_thread_num = false;
   sqlite3_int64 thread_num = 0;

   // Extract "References:" field and check all references against database
   // to see if they are part of a troll thread already.

   std::vector<std::string> & refs = mm.References();
   int numref = refs.size();
   for (int r = 0; r < numref; ++r) {
      std::string ref(refs[r]);
      has_thread_num = check_one_thread(ref, thread_num);
      if (has_thread_num) {
         is_troll_thread = true;
      }
   }

   if (! is_troll_thread) {

      // Extract "From:" field and see if the message sender is a known
      // troll, if so start a new troll thread.

      if (is_a_troll(mm.FromAddress())) {
         is_troll_thread = true;
      }
   }

   if (is_troll_thread) {

      // Extract "Message-ID:" field and record it in database as a known
      // troll thread.

      record_troll_thread(mm.MessageID(), has_thread_num, thread_num);
   }

   if (do_test) {

      // If this was a --test run, return true for nice messages and
      // false for naughty messages. Do not write the filtered message
      // to stdout.

      return ! is_troll_thread;
   }

   // Add X-TrollFilter header to message.

   mm.SetTrollFilter(is_troll_thread);

   // Write modified message to stdout.

   mm.WriteToStdout();

   // If we wrote the filtered message, do not upset the caller by
   // returning an error status.

   return true;
}

static void
add_trolls() {

   // add all of the email addresses in the trolls array to the database
   // trolls table.

   int len = trolls.size();
   for (int i = 0; i < len; ++i) {
      add_one_troll(trolls[i]);
   }
}

static void
delete_trolls() {

   // delete all of the email addresses in the untrolls array from the
   // database trolls table.

   int len = untrolls.size();
   for (int i = 0; i < len; ++i) {
      delete_one_troll(untrolls[i]);
   }
}

static void
add_droppings() {

   // Add all the message IDs in the droppings array
   // as new threads in the droppings table.

   int len = droppings.size();
   for (int i = 0; i < len; ++i) {
      record_troll_thread(droppings[i], false, 0);
   }
}

static void
list_database() {

   // List the trolls in the database and summarize the number of threads
   // and message activity.

   bool first_troll = true;

   std::string cmd("select * from trolls;");
   sqlite3_stmt * sp = NULL;
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      while (sqlite3_step(sp) == SQLITE_ROW) {
         const unsigned char * addr = sqlite3_column_text(sp, 0);
         int addrsize = sqlite3_column_bytes(sp, 0);
         std::string saddr((const char *)addr, addrsize);
         if (first_troll) {
            std::cout << "Trolls in database:" << std::endl;
            first_troll = false;
         }
         std::cout << "   " << saddr << std::endl;
      }
   }
   sqlite3_finalize(sp);
   sp = NULL;
   if (first_troll) {
      std::cout << "There are no troll email addresses in the database."
                << std::endl;
   }

   sqlite3_int64 thread_count = 0;
   cmd = "select count(*) from threads;";
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      if (sqlite3_step(sp) == SQLITE_ROW) {
         thread_count = sqlite3_column_int64(sp, 0);
      }
   }
   sqlite3_finalize(sp);
   sp = NULL;

   sqlite3_int64 dropping_count = 0;
   cmd = "select count(*) from droppings;";
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      if (sqlite3_step(sp) == SQLITE_ROW) {
         dropping_count = sqlite3_column_int64(sp, 0);
      }
   }
   sqlite3_finalize(sp);
   sp = NULL;

   std::cout << "Thread count: " << thread_count << std::endl;
   std::cout << "Message count: " << dropping_count << std::endl;
}

static void
extract_trolladdr(bool add_from) {

   // Extract From: addr from the message read from stdin and either add it
   // or delete it as a troll (based on add_from arg).

   MailMessage mm;
   std::string troll(mm.FromAddress());
   if (add_from) {

      // If we are adding this sender as a troll, be sure to add this
      // message ID as a troll thread as well.

      add_one_troll(troll);
      record_troll_thread(mm.MessageID(), false, 0);
   } else {
      delete_one_troll(troll);
   }
}

static void
extract_trollid() {

   // Extract Message-ID: from the message read from stdin
   // and add it as a new thread in the troll droppings table.

   MailMessage mm;
   record_troll_thread(mm.MessageID(), false, 0);
}

static void
clean_old_threads(unsigned long clean_days) {

   // First delete all the threads with no activity for clean_days days

   std::stringstream ss;
   ss << clean_days;

   std::string cmd("delete from threads where recent < datetime('now', '-");
   cmd.append(ss.str());
   cmd.append(" day');");

   char * errmsg;
   if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Removing old threads error: " << errmsg << std::endl;
      sqlite3_free(errmsg);
   }

   // That leaves a bunch of message IDs that reference threads which no
   // longer exists, so go through the droppings table and remove all the
   // ones with missing thread numbers.

   cmd = "delete from droppings where not exists "
         "(select * from threads "
         "where droppings.threadNum = threads.threadNum);";

   if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Removing old message IDs error: " << errmsg << std::endl;
      sqlite3_free(errmsg);
   }
}

int
main(int argc, char ** argv) {
   unsigned long clean_days = 0;
   bool do_test = false;
   for (int i = 1; i < argc; ++i) {
      if (strncmp(argv[i], "--database=", 11) == 0) {
         if (strcmp(argv[i]+11, "help") == 0) {
            database_help();
            return 0;
         } else {
            database_name.append(argv[i]+11);
         }
      } else if (strncmp(argv[i], "--troll=", 8) == 0) {
         if (strcmp(argv[i]+8, "help") == 0) {
            troll_help();
            return 0;
         } else {
            say_what(PA_MAINT);
            trolls.push_back(std::string(argv[i]+8));
         }
      } else if (strcmp(argv[i], "--trolladdr") == 0) {
         say_what(PA_TROLLADDR);
      } else if (strncmp(argv[i], "--untroll=", 10) == 0) {
         if (strcmp(argv[i]+10, "help") == 0) {
            untroll_help();
            return 0;
         } else {
            say_what(PA_MAINT);
            untrolls.push_back(std::string(argv[i]+10));
         }
      } else if (strcmp(argv[i], "--untrolladdr") == 0) {
         say_what(PA_UNTROLLADDR);
      } else if (strncmp(argv[i], "--dropping=", 11) == 0) {
         if (strcmp(argv[i]+11, "help") == 0) {
            dropping_help();
            return 0;
         } else {
            say_what(PA_MAINT);
            droppings.push_back(std::string(argv[i]+11));
         }
      } else if (strcmp(argv[i], "--trollid") == 0) {
         say_what(PA_TROLLID);
      } else if (strncmp(argv[i], "--clean=", 8) == 0) {
         if (strcmp(argv[i]+8, "help") == 0) {
            clean_help();
            return 0;
         } else {
            char * endp;
            clean_days = strtoul(argv[i]+8, &endp, 10);
            if (*endp != '\0') {
               std::cerr << argv[i]+8 << " is not a valid number." << std::endl;
               return 2;
            }
            say_what(PA_CLEAN);
         }
      } else if (strcmp(argv[i], "--filter") == 0) {
         say_what(PA_FILTER);
      } else if (strcmp(argv[i], "--test") == 0) {
         say_what(PA_FILTER);
         do_test = true;
      } else if (strcmp(argv[i], "--list") == 0) {
         say_what(PA_LIST);
      } else if (strncmp(argv[i], "--help", 6) == 0) {
         help();
         return 0;
      } else {
         std::cerr << "Unrecognized argument: " << argv[i] << std::endl;
         std::cerr << "Try using --help option" << std::endl;
         return 2;
      }
   }

   if (what_to_do == PA_UNSPECIFIED) {
      say_what(PA_FILTER);
   }

   if (database_name.empty()) {
      const char * home = getenv("HOME");
      if ((home == NULL) || (strlen(home) == 0)) {
         std::cerr << "$HOME environment variable is not set." << std::endl;
         return 2;
      }
      database_name.append(home);
      if (database_name[database_name.size()-1] != '/') {
         database_name.append("/");
      }
      database_name.append(".trollfilter");
   }

   open_database();

   int exit_status = 0;

   switch (what_to_do) {
   case PA_MAINT:
      if (trolls.size() > 0) {
         add_trolls();
      }
      if (untrolls.size() > 0) {
         delete_trolls();
      }
      if (droppings.size() > 0) {
         add_droppings();
      }
      break;
   case PA_LIST:
      list_database();
      break;
   case PA_FILTER:
      if (! filter_message(do_test)) {
         exit_status = 3;
      }
      break;
   case PA_TROLLADDR:
      extract_trolladdr(true);
      break;
   case PA_UNTROLLADDR:
      extract_trolladdr(false);
      break;
   case PA_TROLLID:
      extract_trollid();
      break;
   case PA_CLEAN:
      clean_old_threads(clean_days);
      break;
   default:
      std::cerr << "Unimplemented option: "
                << what_name(what_to_do)
                << std::endl;
      exit_status = 2;
      break;
   }

   close_database();

   return exit_status;
}

static void
print_usage_lines(const char ** lines) {
   while (*lines != NULL) {
      std::cerr << *lines++ << std::endl;
   }
}

static const char * database_help_lines[] = {
"--database=file Name the database file to use (default name is",
"$HOME/.trollfilter). The database will be created and initialized if it does",
"not exist.",
NULL
};

static void database_help() {
   print_usage_lines(database_help_lines);
}

static const char * troll_help_lines[] = {
"--troll=address Name an email address to be treated as a troll. This option",
"may be given multiple times to name several trolls. The addresses are added",
"to the database and all mails from these addresses will be considered troll",
"mails.",
NULL
};

static void troll_help() {
   print_usage_lines(troll_help_lines);
}

static const char * untroll_help_lines[] = {
"--untroll=address Name an email address to no longer be treated as a troll.",
"This option may be specified multiple times and all named addresses will be",
"removed from the trolls table.",
NULL
};

static void untroll_help() {
   print_usage_lines(untroll_help_lines);
}

static const char * dropping_help_lines[] = {
"--dropping=message-id Name a message ID to be considered a \"troll",
"dropping\".  Any follow ups to this message ID will also be treated as a",
"troll message.  This option may be given multiple times to name multiple",
"message IDs.",
NULL
};

static void dropping_help() {
   print_usage_lines(dropping_help_lines);
}

static const char * clean_help_lines[] = {
"--clean=days Clean the database of all troll threads that have not had any",
"follow up messages added for the specified number of days.",
NULL
};

static void clean_help() {
   print_usage_lines(clean_help_lines);
}

static const char * help_lines[] = {
"Usage: trollfilter [options...] < email > email",
"",
"trollfilter normally reads an email message on standard in and writes a",
"modified message on standard out with an X-TrollFilter: header added",
"containing the value Yes or No to say if the message is or is not part of a",
"troll thread. Options can modify the standard behavior:",
"",
"--database=file    Give the location of the database file.",
"--troll=address    Name a troll's email address to add to database.",
"--trolladdr        Read mail on stdin and treat From: addr as a troll.",
"--untroll=address  Remove a troll's address from the database.",
"--untrolladdr      Read mail on stdin and remove From: addr as a troll.",
"--dropping=id      Name a message-ID to add to the database.",
"--trollid          Read mail on stdin and add Message-ID: as a troll message.",
"--clean=days       Clean database of troll threads this many days old.",
"--filter           Run as a filter (default action).",
"--test             Read (not write) message, exit status 0 means not a troll.",
"--list             List trolls and threads in the database.",
"--help             Print this message and exit.",
"",
"You can give \"help\" as an argument to any of the options that take",
"arguments to get more information about that option.",
NULL
};

static void help() {
   print_usage_lines(help_lines);
}
